# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)


NAME = 'zhlint'
VERSION = '0.2.1'
AUTHORS = [
    'huntzhan',
]
EMAILS = [
    'programmer.zhx@gmail.com',
]
LICENSE = ''
URL = ''
DESCRIPTION = 'automated checker of chinese document.'
